@@ -13,8 +13,10 @@ class T::Props::Decorator
   Rules = T.type_alias { T::Hash[Symbol, T.untyped] }
   DecoratedInstance = T.type_alias { Object } # Would be T::Props, but that produces circular reference errors in some circumstances
   PropType = T.type_alias { T::Types::Base }
-  PropTypeOrClass = T.type_alias { T.any(PropType, Module) }
+  PropTypeOrClass = T.type_alias { T.any(PropType, T::Module[T.anything]) }
   OverrideRules = T.type_alias { T::Hash[Symbol, {allow_incompatible: T::Boolean}] }
+  DecoratedClassType = T.type_alias { T.all(T::Module[T.anything], T::Props::ClassMethods) }
+  private_constant :DecoratedClassType
 
   class NoRulesError < StandardError; end
 
@@ -38,7 +40,7 @@ class T::Props::Decorator
 
   sig { params(klass: T.untyped).void.checked(:never) }
   def initialize(klass)
-    @class = T.let(klass, T.all(Module, T::Props::ClassMethods))
+    @class = T.let(klass, DecoratedClassType)
     @class.plugins.each do |mod|
       T::Props::Plugin::Private.apply_decorator_methods(mod, self)
     end
@@ -102,7 +104,7 @@ class T::Props::Decorator
   end
 
   # checked(:never) - O(prop accesses)
-  sig { returns(T.all(Module, T::Props::ClassMethods)).checked(:never) }
+  sig { returns(DecoratedClassType).checked(:never) }
   def decorated_class
     @class
   end
@@ -174,7 +176,9 @@ class T::Props::Decorator
     .checked(:never)
   end
   def prop_get(instance, prop, rules=prop_rules(prop))
-    val = instance.instance_variable_get(rules[:accessor_key]) if instance.instance_variable_defined?(rules[:accessor_key])
+    # `instance_variable_get` will return nil if the variable doesn't exist
+    # which is what we want to have happen for the logic below.
+    val = instance.instance_variable_get(rules[:accessor_key])
     if !val.nil?
       val
     elsif (d = rules[:ifunset])
@@ -194,7 +198,9 @@ class T::Props::Decorator
     .checked(:never)
   end
   def prop_get_if_set(instance, prop, rules=prop_rules(prop))
-    instance.instance_variable_get(rules[:accessor_key]) if instance.instance_variable_defined?(rules[:accessor_key])
+    # `instance_variable_get` will return nil if the variable doesn't exist
+    # which is what we want to have happen for the return value here.
+    instance.instance_variable_get(rules[:accessor_key])
   end
   alias_method :get, :prop_get_if_set # Alias for backwards compatibility
 
@@ -203,7 +209,7 @@ class T::Props::Decorator
     params(
       instance: DecoratedInstance,
       prop: Symbol,
-      foreign_class: Module,
+      foreign_class: T::Module[T.anything],
       rules: Rules,
       opts: T::Hash[Symbol, T.untyped],
     )
@@ -222,7 +228,7 @@ class T::Props::Decorator
   sig do
     params(
       name: Symbol,
-      cls: Module,
+      cls: T::Module[T.anything],
       rules: Rules,
       type: PropTypeOrClass
     )
@@ -238,7 +244,9 @@ class T::Props::Decorator
     end
 
     if rules.keys.any? { |k| !valid_rule_key?(k) }
-      raise ArgumentError.new("At least one invalid prop arg supplied in #{self}: #{rules.keys.inspect}")
+      invalid_keys = rules.keys.reject { |k| valid_rule_key?(k) }
+      suffix = invalid_keys.size == 1 ? "" : "s"
+      raise ArgumentError.new("Invalid prop arg#{suffix} supplied in #{self}: #{invalid_keys.inspect}")
     end
 
     if !rules[:clobber_existing_method!] && !rules[:without_accessors] && BANNED_METHOD_NAMES.include?(name.to_sym)
@@ -258,7 +266,9 @@ class T::Props::Decorator
     nil
   end
 
-  SAFE_NAME = T.let(/\A[A-Za-z_][A-Za-z0-9_-]*\z/.freeze, Regexp, checked: false)
+  SAFE_NAME = T.let(/\A[A-Za-z_][A-Za-z0-9_-]*\z/, Regexp, checked: false)
+  # Should be exactly the same as `SAFE_NAME`, but with a leading `@`.
+  SAFE_ACCESSOR_KEY_NAME = T.let(/\A@[A-Za-z_][A-Za-z0-9_-]*\z/, Regexp, checked: false)
 
   # Used to validate both prop names and serialized forms
   sig { params(name: T.any(Symbol, String)).void.checked(:never) }
@@ -269,7 +279,7 @@ class T::Props::Decorator
   end
 
   # This converts the type from a T::Type to a regular old ruby class.
-  sig { params(type: T::Types::Base).returns(Module).checked(:never) }
+  sig { params(type: T::Types::Base).returns(T::Module[T.anything]).checked(:never) }
   private def convert_type_to_class(type)
     case type
     when T::Types::TypedArray, T::Types::FixedArray
@@ -318,7 +328,10 @@ class T::Props::Decorator
 
   sig(:final) { params(name: Symbol).returns(T::Boolean).checked(:never) }
   private def method_defined_on_ancestor?(name)
-    @class.method_defined?(name) && !@class.method_defined?(name, false)
+    (@class.method_defined?(name) || @class.private_method_defined?(name)) &&
+      # Unfortunately, older versions of ruby don't allow the second parameter on
+      # `private_method_defined?`.
+      !@class.method_defined?(name, false) && !@class.private_method_defined?(name, false)
   end
 
   sig(:final) { params(name: Symbol, rules: Rules).void.checked(:never) }
@@ -384,7 +397,7 @@ class T::Props::Decorator
         #
         if sensitivity_and_pii[:pii] && @class.is_a?(Class) && !T.unsafe(@class).contains_pii?
           raise ArgumentError.new(
-            'Cannot include a pii prop in a class that declares `contains_no_pii`'
+            "Cannot define pii prop `#{@class}##{name}` because `#{@class}` is `contains_no_pii`"
           )
         end
       end
@@ -399,8 +412,6 @@ class T::Props::Decorator
 
     # extra arbitrary metadata attached by the code defining this property
 
-    validate_not_missing_sensitivity(name, rules)
-
     # for backcompat (the `:array` key is deprecated but because the name is
     # so generic it's really hard to be sure it's not being relied on anymore)
     if type.is_a?(T::Types::TypedArray)
@@ -476,35 +487,6 @@ class T::Props::Decorator
     end
   end
 
-  # checked(:never) - Rules hash is expensive to check
-  sig { params(prop_name: Symbol, rules: Rules).void.checked(:never) }
-  private def validate_not_missing_sensitivity(prop_name, rules)
-    if rules[:sensitivity].nil?
-      if rules[:redaction]
-        T::Configuration.hard_assert_handler(
-          "#{@class}##{prop_name} has a 'redaction:' annotation but no " \
-          "'sensitivity:' annotation. This is probably wrong, because if a " \
-          "prop needs redaction then it is probably sensitive. Add a " \
-          "sensitivity annotation like 'sensitivity: Opus::Sensitivity::PII." \
-          "whatever', or explicitly override this check with 'sensitivity: []'."
-        )
-      end
-      # TODO(PRIVACYENG-982) Ideally we'd also check for 'password' and possibly
-      # other terms, but this interacts badly with ProtoDefinedDocument because
-      # the proto syntax currently can't declare "sensitivity: []"
-      if /\bsecret\b/.match?(prop_name)
-        T::Configuration.hard_assert_handler(
-          "#{@class}##{prop_name} has the word 'secret' in its name, but no " \
-          "'sensitivity:' annotation. This is probably wrong, because if a " \
-          "prop is named 'secret' then it is probably sensitive. Add a " \
-          "sensitivity annotation like 'sensitivity: Opus::Sensitivity::NonPII." \
-          "security_token', or explicitly override this check with " \
-          "'sensitivity: []'."
-        )
-      end
-    end
-  end
-
   # Create `#{prop_name}_redacted` method
   sig do
     params(
@@ -606,7 +588,7 @@ class T::Props::Decorator
   sig do
     params(
       prop_name: Symbol,
-      prop_cls: Module,
+      prop_cls: T::Module[T.anything],
       rules: Rules,
       foreign: T.untyped,
     )
@@ -649,10 +631,10 @@ class T::Props::Decorator
   #
   # This gets called when a module or class that extends T::Props gets included, extended,
   # prepended, or inherited.
-  sig { params(child: Module).void.checked(:never) }
+  sig { params(child: T::Module[T.anything]).void.checked(:never) }
   def model_inherited(child)
     child.extend(T::Props::ClassMethods)
-    child = T.cast(child, T.all(Module, T::Props::ClassMethods))
+    child = T.cast(child, DecoratedClassType)
 
     child.plugins.concat(decorated_class.plugins)
     decorated_class.plugins.each do |mod|
@@ -739,19 +721,19 @@ class T::Props::Decorator
     result
   end
 
-  sig { params(child: T.all(Module, T::Props::ClassMethods), prop: Symbol).returns(T::Boolean).checked(:never) }
+  sig { params(child: DecoratedClassType, prop: Symbol).returns(T::Boolean).checked(:never) }
   private def clobber_getter?(child, prop)
     !!(child.decorator.method(:prop_get).owner != method(:prop_get).owner &&
        child.instance_method(prop).source_location&.first == __FILE__)
   end
 
-  sig { params(child: T.all(Module, T::Props::ClassMethods), prop: Symbol).returns(T::Boolean).checked(:never) }
+  sig { params(child: DecoratedClassType, prop: Symbol).returns(T::Boolean).checked(:never) }
   private def clobber_setter?(child, prop)
     !!(child.decorator.method(:prop_set).owner != method(:prop_set).owner &&
        child.instance_method("#{prop}=").source_location&.first == __FILE__)
   end
 
-  sig { params(mod: Module).void.checked(:never) }
+  sig { params(mod: T::Module[T.anything]).void.checked(:never) }
   def plugin(mod)
     decorated_class.plugins << mod
     T::Props::Plugin::Private.apply_class_methods(mod, decorated_class)
